shader_type spatial;
render_mode diffuse_burley, specular_schlick_ggx;

uniform bool useplanealbedo = true;
uniform bool useplanenormal = true;
uniform bool useDebug = false;

uniform vec4 planealbedoColor : hint_color;
uniform sampler2D plane_albedo: hint_albedo;
// Ambient occlusion, roughness, metalness
uniform sampler2D plane_normal : hint_normal;
uniform sampler2D snow_albedo: hint_albedo;
uniform sampler2D snow_normal : hint_normal;
uniform float blend_smoothness : hint_range(0.1, 1.0, 0.01) = 0.2;
uniform float snow_range : hint_range(0.1, 1.0, 0.01) = 0.0;
uniform vec3 world_direction = vec3(0.0, 1.0, 0.0);

uniform vec2 planeUV_scale;
uniform vec2 planeUV_offset;

uniform vec2 snowUV_scale;
uniform vec2 snowUV_offset;

//Noise 
uniform float _NoiseScale : hint_range(0.0, 1000.0, 0.01) = 0.0;
uniform float _NoiseOffset : hint_range(0.0, 100.0, 0.01) = 8.0;

//Snow
uniform float _SnowBrightness : hint_range(0.0, 10.0, 0.01) = 1.0;

//Model Repair
uniform bool useModelRepairMode = false;

varying vec2 snowUV;
varying vec2 planeUV;
varying vec3 worldPos;
varying vec3 localCoord;

uniform float _MapScale : hint_range(0.0, 100.0, 0.01) = 1.0;
uniform float _NormalScale : hint_range(0.0, 10.0, 0.01) = 1.0;

/* discontinuous pseudorandom uniformly distributed in [-0.5, +0.5]^3 */
vec3 random3(vec3 c) {
	float j = 4096.0*sin(dot(c,vec3(17.0, 59.4, 15.0)));
	vec3 r;
	r.z = fract(512.0*j);
	j *= .125;
	r.x = fract(512.0*j);
	j *= .125;
	r.y = fract(512.0*j);
	return r-0.5;
}


/* skew constants for 3d simplex functions */
const float F3 =  0.3333333;
const float G3 =  0.1666667;

/* 3d simplex noise */
float simplex3d(vec3 p) {
	 /* 1. find current tetrahedron T and it's four vertices */
	 /* s, s+i1, s+i2, s+1.0 - absolute skewed (integer) coordinates of T vertices */
	 /* x, x1, x2, x3 - unskewed coordinates of p relative to each of T vertices*/
	 
	 /* calculate s and x */
	 vec3 s = floor(p + dot(p, vec3(F3)));
	 vec3 x = p - s + dot(s, vec3(G3));
	 
	 /* calculate i1 and i2 */
	 vec3 e = step(vec3(0.0), x - x.yzx);
	 vec3 i1 = e*(1.0 - e.zxy);
	 vec3 i2 = 1.0 - e.zxy*(1.0 - e);
	 	
	 /* x1, x2, x3 */
	 vec3 x1 = x - i1 + G3;
	 vec3 x2 = x - i2 + 2.0*G3;
	 vec3 x3 = x - 1.0 + 3.0*G3;
	 
	 /* 2. find four surflets and store them in d */
	 vec4 w, d;
	 
	 /* calculate surflet weights */
	 w.x = dot(x, x);
	 w.y = dot(x1, x1);
	 w.z = dot(x2, x2);
	 w.w = dot(x3, x3);
	 
	 /* w fades from 0.6 at the center of the surflet to 0.0 at the margin */
	 w = max(0.6 - w, 0.0);
	 
	 /* calculate surflet components */
	 d.x = dot(random3(s), x);
	 d.y = dot(random3(s + i1), x1);
	 d.z = dot(random3(s + i2), x2);
	 d.w = dot(random3(s + 1.0), x3);
	 
	 /* multiply d by w^4 */
	 w *= w;
	 w *= w;
	 d *= w;
	 
	 /* 3. return the sum of the four surflets */
	 return dot(d, vec4(52.0));
}

/* const matrices for 3d rotation */
const mat3 rot1 = mat3(vec3(-0.37, 0.36, 0.85),vec3(-0.14,-0.93, 0.34),vec3(0.92, 0.01,0.4));
const mat3 rot2 = mat3(vec3(-0.55,-0.39, 0.74), vec3(0.33,-0.91,-0.24),vec3(0.77, 0.12,0.63));
const mat3 rot3 = mat3(vec3(-0.71, 0.52,-0.47),vec3(-0.08,-0.72,-0.68),vec3(-0.7,-0.45,0.56));

/* directional artifacts can be reduced by rotating each octave */
float simplex3d_fractal(vec3 m) {
    return   0.5333333*simplex3d(m*rot1)
			+0.2666667*simplex3d(2.0*m*rot2)
			+0.1333333*simplex3d(4.0*m*rot3)
			+0.0666667*simplex3d(8.0*m);
}

vec3 UnpackScaleNormal(vec3 packNormal, float bumpScale){
	return (packNormal * 2.0 - vec3(1.0)) * bumpScale;
}

//Snow Model Normal mode
void CalculateSnowByModelNormal(vec4 plane_albedotex, vec4 snow_albedotex, vec3 plane_normalmap, vec3 snow_normalmap, vec3 world_tangent, vec3 world_binormal, vec3 world_normal, 
								float noise, vec3 modelNormal, inout vec3 finalSnowAlbedo, inout vec3 finalSnowNormalMap) {
	float mix_factor = 0.0;
	float addends = 0.0;

	vec3 snow_world_normalmap = vec3(0, 0, 0);

	// textures are from 0 to 1, but we want the normals to be able to go negative.
	vec3 unpacked_normal = UnpackScaleNormal(snow_normalmap, _NormalScale);        //snow_normalmap * 2.0 - vec3(1.0);
	unpacked_normal.z = sqrt(max(0.0, 1.0 - dot(unpacked_normal.xy, unpacked_normal.xy)));

	// Normal mixing taken directly from Godot's spatial shader (drivers/gles3/shaders/spatial.glsl)
	snow_world_normalmap = normalize(unpacked_normal.y * world_tangent + unpacked_normal.x * world_binormal + world_normal * unpacked_normal.z);

	// the more the world-space normal is aligned to the given vector,
	// the more we want to show the second texture.
	mix_factor += clamp(
		dot(snow_world_normalmap, world_direction)
		, 0.0, 1.0);
	addends += 1.0;

	mix_factor /= addends;
	// This part makes the mix factor more crisp than a bland mix between zero and one.
	// threshold allows us to cut away more of the second texture, while smoothness defines
	// the crispiness of the edges of the blending. The smoother the blending, the more diluted
	// and blended the textures will look like.
	// It would be interesting to use the displacement map of the second texture as a small
	// boost to the mix_factor, so that blending can follow the shapes in the second texture.
	// Here are the textures: https://cc0textures.com/view?id=Moss003
	// Have fun!
	mix_factor = smoothstep(0.0, blend_smoothness, mix_factor - snow_range);

	if (useplanenormal) {
		finalSnowNormalMap = mix(plane_normalmap, snow_normalmap, mix_factor);
	}
	else {
		finalSnowNormalMap = mix(modelNormal, snow_normalmap, mix_factor);
	}
	//NORMALMAP = mix(plane_normalmap, snow_normalmap, mix_factor);

	if (useplanealbedo) {
		
		if(_NoiseScale == 0.00){
			finalSnowAlbedo = clamp(mix(plane_albedotex, snow_albedotex, mix_factor*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}else{
			finalSnowAlbedo = clamp(mix(plane_albedotex, snow_albedotex, mix_factor*(noise*_SnowBrightness)).rgb, vec3(0.0), vec3(1.0));
		}
		
		//ALBEDO = clamp(mix(plane_albedotex, snow_albedotex*noise, mix_factor).rgb, vec3(0.0), vec3(1.0));
	}
	else {
		
		if(_NoiseScale == 0.00){
			finalSnowAlbedo = clamp(mix(planealbedoColor, snow_albedotex, mix_factor*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}else{
			finalSnowAlbedo = clamp(mix(planealbedoColor, snow_albedotex, mix_factor*(noise*_SnowBrightness)).rgb, vec3(0.0), vec3(1.0));			
		}
			
	}
	
		//Debug information
	if(useDebug){
		finalSnowAlbedo = vec3(mix_factor);
	}
	
}

//Snow Model Normal mode
void CalculateSnowByModelRepair(vec4 plane_albedotex, vec4 snow_albedotex, vec3 plane_normalmap, vec3 snow_normalmap, vec3 world_tangent, vec3 world_binormal,
	float noise, vec3 modelNormal, inout vec3 finalSnowAlbedo, inout vec3 finalSnowNormalMap) {
	float mix_factor = 0.0;
	float addends = 0.0;

	//Recalculate world space normals for model
	vec3 world_normal = normalize(cross(dFdy(worldPos.xyz), dFdx(worldPos.xyz))) * -1.0;

	//Recalculate UV for model
	// Blending factor of triplanar mapping
	vec3 bf = normalize(abs(world_normal));
	bf /= dot(bf, vec3(1.0, 1.0, 1.0));

	// Triplanar mapping
	vec2 tx = worldPos.yz * _MapScale;
	vec2 ty = worldPos.zx * _MapScale;
	vec2 tz = worldPos.xy * _MapScale;

	// Base color
	vec4 cx = texture(snow_albedo, tx) * bf.x;
	vec4 cy = texture(snow_albedo, ty) * bf.y;
	vec4 cz = texture(snow_albedo, tz) * bf.z;
	vec4 snowBaseColor = (cx + cy + cz);

	vec3 snow_world_normalmap = vec3(0, 0, 0);
	vec3 plane_world_normalmap = vec3(0, 0, 0);

	// Normal map
	vec4 nx = texture(snow_normal, tx) * bf.x;
	vec4 ny = texture(snow_normal, ty) * bf.y;
	vec4 nz = texture(snow_normal, tz) * bf.z;
	vec3 resnownormalmap = UnpackScaleNormal(nx.xyz + ny.xyz + nz.xyz, _NormalScale);

	//Snow NormalMap
	// textures are from 0 to 1, but we want the normals to be able to go negative.
	vec3 unpacked_SnowNormal = resnownormalmap.rgb; //* 2.0 - vec3(1.0);
	unpacked_SnowNormal.z = sqrt(max(0.0, 1.0 - dot(unpacked_SnowNormal.xy, unpacked_SnowNormal.xy)));

	// Normal mixing taken directly from Godot's spatial shader (drivers/gles3/shaders/spatial.glsl)
	snow_world_normalmap = normalize(unpacked_SnowNormal.y * world_tangent + unpacked_SnowNormal.x * world_binormal + world_normal * unpacked_SnowNormal.z);

	// the more the world-space normal is aligned to the given vector,
	// the more we want to show the second texture.
	mix_factor += clamp(
		dot(snow_world_normalmap, world_direction)
		, 0.0, 1.0);
	addends += 1.0;

	mix_factor /= addends;
	// This part makes the mix factor more crisp than a bland mix between zero and one.
	// threshold allows us to cut away more of the second texture, while smoothness defines
	// the crispiness of the edges of the blending. The smoother the blending, the more diluted
	// and blended the textures will look like.
	// It would be interesting to use the displacement map of the second texture as a small
	// boost to the mix_factor, so that blending can follow the shapes in the second texture.
	// Here are the textures: https://cc0textures.com/view?id=Moss003
	// Have fun!
	mix_factor = smoothstep(0.0, blend_smoothness, mix_factor - snow_range);

	if (useplanenormal) {
		finalSnowNormalMap = mix(plane_normalmap, snow_normalmap, mix_factor);
	}
	else {
		finalSnowNormalMap = mix(world_normal, snow_normalmap, mix_factor);
	}
	//NORMALMAP = mix(plane_normalmap, snow_normalmap, mix_factor);

	if (useplanealbedo) {
		
		if(_NoiseScale == 0.00){
			finalSnowAlbedo = clamp(mix(plane_albedotex, snowBaseColor, mix_factor*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}else{
			finalSnowAlbedo = clamp(mix(plane_albedotex, snowBaseColor, mix_factor*noise*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}

	}
	else {
		
		if(_NoiseScale == 0.00){
			finalSnowAlbedo = clamp(mix(planealbedoColor, snowBaseColor, mix_factor*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}else{
			finalSnowAlbedo = clamp(mix(planealbedoColor, snowBaseColor, mix_factor*noise*_SnowBrightness).rgb, vec3(0.0), vec3(1.0));
		}

	}
	
	//Debug information
	if(useDebug){
		finalSnowAlbedo = vec3(mix_factor);
	}
	
}

void vertex() {
	snowUV = UV * snowUV_scale.xy * _MapScale + snowUV_offset.xy;
	planeUV = UV * planeUV_scale.xy + planeUV_offset.xy;
	worldPos = (WORLD_MATRIX * vec4(VERTEX, 1.0)).xyz;
}

void fragment(){
	// sample all the textures
	vec4 plane_albedotex = texture(plane_albedo, planeUV)*planealbedoColor;
	vec4 snow_albedotex = texture(snow_albedo, snowUV);
	vec3 plane_normalmap = texture(plane_normal, planeUV).rgb;
	vec3 snow_normalmap = texture(snow_normal, snowUV).rgb;

	// Binormal and tangent vectors are what allows us to transform the
	// 2D normalmaps into something that can interact with 3D space
	vec3 world_tangent = (CAMERA_MATRIX * vec4(TANGENT, 0.0)).xyz;
	vec3 world_binormal = (CAMERA_MATRIX * vec4(BINORMAL, 0.0)).xyz;
	vec3 world_normal = (CAMERA_MATRIX * vec4(NORMAL, 0.0)).xyz;
	
	//3D Noise Calculation
	vec2 p = VIEWPORT_SIZE;
	float noise = simplex3d_fractal(worldPos*_NoiseScale + _NoiseOffset);
	noise = 0.5 + 0.5*noise;
	noise *= smoothstep(0.0, 0.005, abs(0.6-p.x)); // hello, iq :)

	vec3 finalSnowAlbedo = vec3(0.0);
	vec3 finalSnowNormalMap = vec3(0.0);

	if (useModelRepairMode) {
		CalculateSnowByModelRepair(plane_albedotex, snow_albedotex, plane_normalmap, snow_normalmap, world_tangent, world_binormal,
								   noise, NORMAL, finalSnowAlbedo, finalSnowNormalMap);
	}
	else {
		CalculateSnowByModelNormal(plane_albedotex, snow_albedotex, plane_normalmap, snow_normalmap, world_tangent, world_binormal, world_normal,
								   noise, NORMAL, finalSnowAlbedo, finalSnowNormalMap);
	}

	ALBEDO = finalSnowAlbedo;
	NORMALMAP = finalSnowNormalMap;
}

